; Copyright  2004 Ben J. Jackson <ben@ben.com>
; May be freely used, modified and distributed,
; provided this notice remains intact.  This software
; is provided "as is" without express or implied warranty.

; EXAMPLE:
;
; include "18fmul.inc"
; cblock 0x80 ; each arg must be contiguous,
;             ; but the args don't have to be adjacent
;  ARGA:6 ; first term and destination, wide enough to hold result
;         ; MSB in ARGA, LSB in ARGA+n
;  ARGB:2 ; second term (will be unchanged)
;  TEMPA:4 ; copy of first term, only big enough to hold term
;  LOOP
; endc
; mkmul foo, ARGA, ARGB, TEMPA, d'32', d'16', UNSIGNED
;
; That makes a function foo32x16u which multiplies ARGA and ARGB
; and puts the result (as wide as the sum of the widths of both
; arguments, 6 bytes in this case) in ARGA.
;
; Change the last argument to SIGNED to treat ARGA, ARGB and the
; result in ARGA all as signed.
;
; Since the function always copies ARGA it will be slightly more
; efficient to make the first argument the short one.
;
; There are hand-coded multiply functions for some short combinations
; of terms that reduce or avoid the copy to TEMP (especially in the
; unsigned case).  If you need one you can use this macro to generate
; a starting point and hand bum from there.
;
SIGNED equ 1
UNSIGNED equ 0

mkmul macro name, aa, bb, temp, abits, bbits, signed
	local awords = (abits + 7) / 8
	local bwords = (bbits + 7) / 8
	if (abits != awords * 8) || (bbits != bwords * 8)
		error "18fmul has no optimizations for odd bit length terms."
	endif
	if signed
name#v(abits)x#v(bbits)s:
	else
name#v(abits)x#v(bbits)u:
 	endif	
	; simplify macro by copying all of A even in unsigned case
	local i = awords
	while i > 0
	local i = i - 1
	movff aa+i, temp+i
	endw
	; first order of business is to do some muls which
	; reinitialize aa.  if the result length is odd, then
	; the last byte of the output will just be clrf'd.

	; Make a poor-man's 2d array of product terms and mark
	; them all unused
	local i = 0
	local j
	while i < awords
	 local j = 0
	 while j < bwords
	   local mul#v(i)x#v(j)used = 0
	 local j = j + 1
	 endw
	local i = i + 1
	endw
	
	; how many product terms remain
	local needed = awords * bwords

	; create a coverage map for the output bytes and
	; mark them all unused for the first pass (where
	; results are copied instead of added)
	local i = 0
	while i < (awords + bwords)
	   local mul#v(i)cov = 0
	local i = i + 2
	endw

	local i = 0
	while i < awords
		local j = 0
		while j < bwords
			if (((i+j) & 1) == 0)
	   			if (mul#v(i+j)cov == 0)
					movf temp+i, W
					mulwf bb+j
					movff PRODL, aa+i+j+1 
					movff PRODH, aa+i+j 
					local mul#v(i)x#v(j)used = 1
					local mul#v(i+j)cov = 1
					local needed = needed - 1
				endif
			endif
			local j = j + 1
		endw
	local i = i + 1
	endw
	; if there are an odd number, the first product term
	; below should pick it up, make a note!
	local oddresid = ((awords + bwords) & 1)
	; now compute all of the other products and add them in
	; use a 'painting' approach, alternating odd and even
	; pairs of output bytes until coverage is complete
	local pass = 1
	local wisclr = 0
   while needed > 0
	; k will be the target lsb of the mul (ie prodl's dest)
	local k = awords + bwords - 1 - (pass & 1)
	local first = 1
	local hitit = -1
	while k >= 1
		local i = awords
		while i > 0
	 		local i = i - 1
	 		local j = bwords
	 		while j > 0
		  		local j = j - 1
		  		if (hitit != k && (i+j+1 == k) && mul#v(i)x#v(j)used == 0)
		  			local wisclr = 0
	    			movf temp+i, W
	    			mulwf bb+j
					if oddresid
						local oddresid = 0
						movff PRODL, aa+i+j+1
						movf PRODH, W
						addwf aa+i+j, F
						local first = 0
					else
						movf PRODL, W
						if first
			 				addwf aa+i+j+1, F
			 				local first = 0
						else
							addwfc aa+i+j+1, F
						endif
						movf PRODH, W
						addwfc aa+i+j, F
					endif
				   local mul#v(i)x#v(j)used = 1
				   local needed = needed - 1
				   local hitit = k
				endif
			endw
		endw
		if (hitit != k && first == 0)
			; don't even know if we need this, but if
			; we skip a product term we need to complete
			; the carry chain
			if wisclr == 0
				clrf WREG
				local wisclr = 1
			endif
			addwfc aa+k, F
			addwfc aa+k-1, F
		endif
		local k = k - 2
	endw
	; on odd passes bring the carry chain up to aa+0
	if (k == 0 && first ==0)
		if wisclr == 0
			clrf WREG
			local wisclr = 1
		endif
		addwfc aa, F
	endif
	local pass = pass + 1
	endw
	
	; handle signedness
	if signed
	btfss temp, 7
	bra name#v(abits)x#v(bbits)n#v(signed)
	local i = bwords - 1
	movf bb+i, W
	subwf aa+i, F
	while i > 0
	local i = i - 1
	movf bb+i, W
	subwfb aa+i, F
	endw
name#v(abits)x#v(bbits)n#v(signed):
	btfss bb, 7
	return
	local i = awords - 1
	movf temp+i, W
	subwf aa+i, F
	while i > 0
	local i = i - 1
	movf temp+i, W
	subwfb aa+i, F
	endw
	endif  
	return
	endm
